/************************************************************************
* \brief: Scientific Library                                            *
*																		*
* (c) copyright by Jrn Fischer											*
*                                                                       *
*        ***  ITS PROHIBITED TO DISTRIBUTE THIS SOURCE!  ***            *
*        ***************************************************            *
*																		* 
* @autor: Prof.Dr.Jrn Fischer											*
* @email: j.fischer@hs-mannheim.de										*
*                                                                       *
* @file : VirtualMachine.h                                                      *
*************************************************************************/

#ifndef VIRTUAL_MACHINE
#define VIRTUAL_MACHINE


#define MAX_THREAD 10
#define MEM_SIZE 0xFFFFF
#define STACK_SIZE 0xFFFF

enum XCMD{
	CAST,
	MOVE,
	MUL,
	DIV,
	ADD,
	SUB,
	AND,
	OR,
	NOT,
	HIG,
	EQU,
	JIT,
	JSR,
	JMP,
	RET,
	PUSH,
	POP,
	MOD
};


union mem{
	float FLOAT;
	unsigned int INT;
};
/**********************************************************************************************
\brief Simulates a Processor with the following commands:
LOAD 10 // LSB 0, 31 Bit Float
CAST Ri,Rj // LSB 1, 5 Bit Befehl, 8 Bit j, 8 Bit i // Ri = (int)Ri
MOVE Ri,(Rj) // LSB 1, 5 Bit Befehl, 8 Bit j, 8 Bit i, 2 Bit Ri/Rj as pointer,
MOVE (Ri), Rj
MOVE (Ri),(Rj)
MOVE Ri , Rj
AND Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
OR Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
MUL Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
DIV Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
SUB Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
ADD Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
HIG Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
EQU Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
NOT Ri,Rj // LSB 1, 5 Bit Befehl, 8 Bit j, 8 Bit i
JIT Ri label // LSB 1, 5 Bit Befehl, 26 Bit Label if (regi) goto label
JSR label // LSB 1, 5 Bit Befehl, 26 Bit Label JumptoSubRoutine
JMP label // LSB 1, 5 Bit Befehl, 26 Bit Label Jump
RET // LSB 1, 5 Bit Befehl
PUSH // LSB 1, 5 Bit Befehl
POP // LSB 1, 5 Bit Befehl
MOD Ri,Rj,Rk // LSB 1, 5 Bit Befehl, 8 Bit k, 8 Bit j ,8 Bit i
**********************************************************************************************/
class VirtualMachine
{
public:
	char text[100];
	float reg[256];
	mem memory[MEM_SIZE+1];
	int prgCNT[MAX_THREAD];
	int idx0,idx1,idx2;
	int adress;
	float stack[MAX_THREAD][STACK_SIZE];
	int stackPtr[MAX_THREAD];
	float value1,value2;
	int prg,opcode,cycle;

	int HW_THREAD_ADRESS;

	int actualThread;
	int numberOfExistingThreads;
	int cyclesAfterThreadChanges;

	int threadEnable[MAX_THREAD];
	int threadAddress[MAX_THREAD];


public:

	void hw_init()
	{
		HW_THREAD_ADRESS = (int)(memory[4].FLOAT);

	    actualThread = (int)(memory[HW_THREAD_ADRESS & MEM_SIZE].FLOAT);
		numberOfExistingThreads= (int)(memory[(HW_THREAD_ADRESS+1)& MEM_SIZE].FLOAT);
		cyclesAfterThreadChanges= (int)(memory[(HW_THREAD_ADRESS+2)& MEM_SIZE].FLOAT);

		printf("actualThread = %d\n;numberOfExistingThreads=%d;\ncyclesAfterThreadChanges=%d;\n",actualThread,numberOfExistingThreads,cyclesAfterThreadChanges);
		for (int t=0;t<numberOfExistingThreads;t++){
			threadEnable[t] = (int)(memory[(HW_THREAD_ADRESS+3+t*2)& MEM_SIZE].FLOAT);
			threadAddress[t]= (int)(memory[(HW_THREAD_ADRESS+4+t*2)& MEM_SIZE].FLOAT);
			printf("Adress[%d] = %d;\n",t,threadAddress[t]);
			prgCNT[t]       = threadAddress[t];
			stackPtr[t]     = 0;
		}
		cycle    = 0;

	}
	void simulateStep()
	{

		// BIN(0000 0000  0000 0000  0000 0000  0000 0001) = (opcode & 0x00000001)==0  (lowest bit is equal to zero)
		// BIN(0000 0000  0000 0000  0000 0000  0011 1110) = 0x0000001E opcode
		// BIN(0000 0000  0000 0000  0000 0000  0100 0000) = 0x40 Adress(Reg[idx1])
		// BIN(0000 0000  0000 0000  0000 0000  1000 0000) = 0x80 Adress(Reg[idx2])
		// BIN(0000 0000  0000 0000  1111 1111  0000 0000) = 0x0000FF00 idx0
		// BIN(0000 0000  1111 1111  0000 0000  0000 0000) = 0x00FF0000 idx1
		// BIN(1111 1111  0000 0000  0000 0000  0000 0000) = 0xFF000000 idx2
		// BIN(1111 1111  1111 1111  1111 1111  1100 0000) = 0xFFFFFC00 label
		do{ cycle++;
			prg = memory[(prgCNT[actualThread]) & MEM_SIZE].INT;
			opcode = (prg & 0x0003E)>>1;
			idx0 =   (prg & (0x0000FF00)) >> 8;
			idx1 =   (prg & (0x00FF0000)) >> 16;
			idx2 =   (prg & (0xFF000000)) >> 24;
			adress = (prg & (0xFFFFFFC0)) >> 6; // 26 Bit Adress

			printf("%6d : ",prgCNT[actualThread]);
			if ((prg & 0x01)==0){ // LOAD
				reg[0] = memory[prgCNT[actualThread] & MEM_SIZE].FLOAT;
				printf("LOAD %f",memory[prgCNT[actualThread]& MEM_SIZE].FLOAT);
			}
			else{
				switch(opcode){
				case(CAST): {reg[idx0] = (int)reg[idx1]        ; printf("CAST R%d,R%d",idx0,idx1);break;}
				case(MUL) : {reg[idx0] = reg[idx1] * reg[idx2] ; printf("MUL  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}	 
				case(DIV) : {reg[idx0] = reg[idx1] / reg[idx2] ; printf("DIV  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}
				case(ADD) : {reg[idx0] = reg[idx1] + reg[idx2] ; printf("ADD  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}
				case(SUB) : {reg[idx0] = reg[idx1] - reg[idx2] ; printf("SUB  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}
				case(AND) : {reg[idx0] = (int)reg[idx1] & (int)reg[idx2] ; printf("AND  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}
				case(OR)  : {reg[idx0] = (int)reg[idx1] | (int)reg[idx2] ; printf("OR  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}
				case(NOT) : {reg[idx0] = (int)(reg[idx1])==0?1:0  ; printf("NOT  R%d,R%d          R%d=%f  R%d=%f",idx0,idx1,idx0,reg[idx0],idx1,reg[idx1]);break;}
				case(HIG) : {reg[idx0] = reg[idx1] > reg[idx2] ; printf("HIG  R%d,R%d,R%d       R%d=%f  R%d=%f  R%d=%f",idx0,idx1,idx2,idx0,reg[idx0],idx1,reg[idx1],idx2,reg[idx2]);break;}
				case(EQU) : {reg[idx0] = reg[idx1] == reg[idx2]; printf("EQU  R%d,R%d,R%d",idx0,idx1,idx2);break;}
				case(JIT) : {if (reg[0]){
					prgCNT[actualThread] = adress;
							}else{
								prgCNT[actualThread]++;
							}
							printf("JIT %d",adress);break;}
				case(JSR) : {stack[actualThread][stackPtr[actualThread]++] = prgCNT[actualThread]; 
					for (int t=0;t<256;t++){
						stack[actualThread][stackPtr[actualThread]++] = reg[t];
					}
					prgCNT[actualThread] = adress; printf("JSR  %d    stack=%d",prgCNT[actualThread],stackPtr[actualThread]);break;}
				case(JMP) : {prgCNT[actualThread] = adress; printf("JMP  %d",prgCNT[actualThread]);break;}
				case(RET) : {     
					for (int t=255;t>=0;t--){
						reg[t] = stack[actualThread][--stackPtr[actualThread]];
					}
					prgCNT[actualThread]=stack[actualThread][--stackPtr[actualThread]]+1;

					//for (int t=0;t<20;t++){printf("mem[%d]=%d\n",t,memory[t].INT);}

					printf("RET    stack= %d",stackPtr[actualThread]);break;}
				case(MOVE): {sprintf(text,"R%d",idx1);
					float value = reg[idx1];	
					if (prg & (0x80)) {
						value = memory[((int)reg[idx1])&MEM_SIZE].FLOAT;
						sprintf(text,"(R%d)",idx1);
					}
					if (prg & (0x40)) {
						memory[((int)reg[idx0])&MEM_SIZE].FLOAT = value;
						printf("MOVE  (R%d),%s        R%d=%f, R%d=%f",idx0,text,idx0,reg[idx0],idx1,reg[idx1]);
					}else{
						reg[idx0] = value;
						printf("MOVE  R%d,%s          R%d=%f, R%d=%f",idx0,text,idx0,reg[idx0],idx1,reg[idx1]);
					}

					break;}
				case(PUSH): {
						printf("PUSH\n");
						for (int t=100;t<255;t++){
							stack[actualThread][stackPtr[actualThread]++] = reg[t];
						}
						break;
					} 
				case(POP): {
						printf("POP");
						for (int t=254;t>=100;t--){
									reg[t] = stack[actualThread][--stackPtr[actualThread]];
						}
						break;
					}
				case(MOD): {
						printf("MOD...\n");
						reg[idx0] = (int)reg[idx1] % (int)reg[idx2] ;
						break;
					}
				}
			}
			if (opcode!= JIT && opcode!=JSR && opcode!=RET && opcode!= JMP && (prg & 0x01)==1){
				prgCNT[actualThread]++;
			}
			if ((prg & 0x01)==0){
				prgCNT[actualThread]++;
			}
			if (cycle % 300000 == 0){
				display();
				glutMainLoopEvent();
			}			printf("\n");
			if (cycle%20==0){
				system("pause");
			}
		}while( cycle < cyclesAfterThreadChanges );
		
		// TODO: threadwechsel
	}


};
#endif//VIRTUAL_MACHINE